﻿using System;
using System.Collections.Generic;
using System.Threading;
using System.Linq;
using System.Text;
using RS232Port;

namespace NavigationController {
	public class NavigationStrategy : SerialStrategy {

		public event updateStatusEventHandler updateResetEvent;

		NavigationData _data;
		public NavigationData Data {
			get { return _data; }
		}

		public NavigationStrategy(string portName, double magDec) {
			_data = new NavigationData(magDec);
			serialPortName = portName;
			initializePort();
			_watchdogTimer.Start();
		}

		private void initializePort() {
			_rs232 = new RS232MainForm();
			_rs232.updateErrorEvent += updateError;
			_rs232.receiveDataEvent += receiveData;
			_rs232.setSilentModeEvent += updateSilentMode;
			_rs232.Show();
			_rs232.Hide();
			_rs232.Port.ReceivedBytesThreshold = 256;
			_rs232.DebugMode = RS232MainForm.debugTerminalMode.silent;
			_rs232.Port.BaudRate = 19200;
			_rs232.updatePortSettings();
			_rs232.Text = "Navigation Controller Serial Port";
			_rs232.Port.PortName = serialPortName;
			_status = _rs232.initializeComPort();
			resetNavigationController();
		}

		public void resetNavigationController() {
			if (_rs232.Port.IsOpen) {
				_rs232.Port.DtrEnable = true;
				_rs232.Port.RtsEnable = true;
				Thread.Sleep(100);
				_rs232.Port.RtsEnable = false;
				_rs232.Port.DtrEnable = false;
			}
		}

		public void writeToNavigationController(NavigationCommand command) {
			_rs232.transmit(command.command.ToString());
		}

		private void receiveData() {
			parsSerialData(_rs232.SerialData);
			if (_data.CompassHeadList.Count > 5) {
				_data.CompassHeadList.Sort();
				_data.CompassHeading = _data.getRealCompassHeading
					(_data.CompassHeadList[_data.CompassHeadList.Count / 2]);
				updateUi();
				updateData();
				clearDataLists();
				_watchdogTimer.Reset();
				_watchdogTimer.Start();
			}
		}

		//part of the strategy class is the parsing algorithm 
		//for the navigation controller's serial data.
		public void parsSerialData(string buffer) {
			try {
				_data.NavigationBuffer = buffer;
				string s = "";
				char c = '0';
				buffer = buffer.Remove(0, buffer.IndexOf(";") + 1);
				buffer = buffer.Remove(buffer.LastIndexOf(";") + 1
					, buffer.Length - (buffer.LastIndexOf(";") + 1));
				while (buffer.Length > 0) {
					s = buffer.Substring(buffer.IndexOf(":") - 1, buffer.IndexOf(";"));
					c = s[0];
					s = s.Remove(0, 2);
					switch (c) {
						case '0':
							_data.Status = s;
							updateStatus();
							_rs232.Port.DiscardInBuffer();
							break;
						case 'a':
							//_data.EncoderCounts = int.Parse(s);
							break;
						case 'b':
							if (double.TryParse(s, out _outDouble)) {
								_data.LeftIrList.Add(_outDouble);
								_data.LeftIr = _outDouble;
							}
							break;
						case 'c':
							if (double.TryParse(s, out _outDouble)) {
								_data.RightIrList.Add(_outDouble);
								_data.RightIr = _outDouble;
							}
							break;
						case 'd':
							if (double.TryParse(s, out _outDouble)) {
								_data.SystemBattList.Add(_outDouble);
								_data.SystemBattery = _outDouble;
							}
							break;
						case 'e':
							if (double.TryParse(s, out _outDouble)) {
								_data.PlatformBattList.Add(_outDouble);
								_data.PlatformBattery = _outDouble;
							}
							break;
						case 'f':
							if (double.TryParse(s, out _outDouble)) {
								if (_outDouble < 3600 && _outDouble >= 0) {
									_outDouble += _data.magneticDeclination * 10;
									_data.CompassHeadList.Add(_outDouble);
								}
							}
							break;
						case 'g':
							if (double.TryParse(s, out _outDouble)) {
								_data.XAccList.Add(_outDouble);
								_data.XAcc = _outDouble;
							}
							break;
						case 'h':
							if (double.TryParse(s, out _outDouble)) {
								_data.YAccList.Add(_outDouble);
								_data.YAcc = _outDouble;
							}
							break;
						case 'i':
							if (double.TryParse(s, out _outDouble)) {
								_data.ZAccList.Add(_outDouble);
								_data.ZAcc = _outDouble;
							}
							break;
						case 'j':
							writeToNavigationController(
								NavigationCommands.cmdArray[
								(int)NavigationCommands
								.commands.resetEncoderLedOff]);
							if (updateResetEvent != null)
								updateResetEvent.Invoke();

							break;
						default:
							break;
					}
					buffer = buffer.Substring(buffer.IndexOf(";") + 1
						, buffer.Length - (buffer.IndexOf(";") + 1));
				}
			} catch (Exception e) {
				updateError(e);
				_error = e.Message + "\r\n" + e.StackTrace;
			}
		}

		public void clearDataLists() {
			_data.LeftIrList.Clear();
			_data.RightIrList.Clear();
			_data.SystemBattList.Clear();
			_data.PlatformBattList.Clear();
			_data.CompassHeadList.Clear();
			_data.XAccList.Clear();
			_data.YAccList.Clear();
			_data.ZAccList.Clear();
		}

	}
}
